require "draw_primitive.rb"


# --------------------------------------------------------
#  line mixing
# --------------------------------------------------------
module LineCrossCheck
  def crossbase(line1, line2)
    result = Hash.new
    l1p1 = line1.get_point(0)
    l1p2 = line1.get_point(1)
    l2p1 = line2.get_point(0)
    l2p2 = line2.get_point(1)
    det = (l2p2.x - l2p1.x)*(l1p2.y - l1p1.y) - (l1p2.x - l1p1.x)*(l2p2.y - l2p1.y)
    if det == 0 then
      return nil
    end
    dx = l2p1.x - l1p1.x
    dy = l2p1.y - l1p1.y
    t1 = ((l2p2.x - l2p1.x)*dy - (l2p2.y - l2p1.y)*dx) / det
    t2 = ((l1p2.x - l1p1.x)*dy - (l1p2.y - l1p1.y)*dx) / det
    t1 = 0 if t1 < 0.00000001 && t1 > 0
    t2 = 0 if t2 < 0.00000001 && t1 > 0
    t1 = 1.0 if t1 > 0.99999999 && t1 < 1.0
    t2 = 1.0 if t2 > 0.99999999 && t1 < 1.0
    result['t1'] = t1
    result['t2'] = t2
    return result
  end

  def crosspoint(line1, line2)
    t = crossbase(line1, line2)
    l1p1 = line1.get_point(0)
    l1p2 = line1.get_point(1)
    l2p1 = line2.get_point(0)
    l2p2 = line2.get_point(1)
    crosspoint = PPoint.new
    crosspoint.x = l1p1.x + (l1p2.x - l1p1.x) * t['t1']
    crosspoint.y = l1p1.y + (l1p2.y - l1p1.y) * t['t1']
    crosspoint.z = 0
    return crosspoint
  end

  def cross?(line1, line2)
    t = crossbase(line1, line2)
    if t == nil then
      return false
    end
    return ((t['t1'] > 0) && (t['t1'] < 1.0) && (t['t2'] > 0) && (t['t2'] < 1.0))
  end

  def cross_with_top?(line1, line2)
    t = crossbase(line1, line2)
    if t == nil then
      return false
    end
    return ((t['t1'] >= 0) && (t['t1'] <= 1.0) && (t['t2'] >= 0) && (t['t2'] <= 1.0))
  end

  def online?(line, point)
    result = false
    l1p1 = line.get_point(0)
    l1p2 = line.get_point(1)
    result=true if l1p1.equal?(point)
    result=true if l1p2.equal?(point)
    v1 = PPoint.new(l1p2.x-l1p1.x, l1p2.y-l1p1.y)
    v2 = PPoint.new(point.x-l1p1.x, point.y-l1p1.y)
    inner = v1.inner_product(v2)
    if (inner > 0.999999) && (v1.length >= v2.length) then
      result = true
    end
    return result
  end

end


# --------------------------------------------------------
#  triangle marge work
# --------------------------------------------------------

class TriangleCollision
  include LineCrossCheck

  def initialize(t1)
    @tri1 = t1
    @div1 = []
    @div2 = []
    @epoint1 = []
    @epoint2 = []
    @epoint3 = []
  end

  attr_accessor :div1
  attr_accessor :div2

  def marge(tri2)
    @div1.clear
    @div2.clear
    divide_crossline(tri2)
  end

  def build_div1()
    @div1.clear
    subdivision
  end

  # ------------------------------
  # --- tool method
  def include?(tri1, tri2)
    p1 = tri2.get_point(0)
    if (tri1.include?(p1) == false) && (onedge?(tri1, p1) == false) then
      return false
    end
    p2 = tri2.get_point(1)
    if (tri1.include?(p2) == false) && (onedge?(tri1, p2) == false) then
      return false
    end
    p3 = tri2.get_point(2)
    if (tri1.include?(p3) == false) && (onedge?(tri1, p3) == false) then
      return false
    end
    return true    
  end

  def onedge?(tri, point)
    line = PLine.new
    line.set_point(0, tri.get_point(0))
    line.set_point(1, tri.get_point(1))
    if online?(line, point) then
      return true
    end
    line.set_point(0, tri.get_point(1))
    line.set_point(1, tri.get_point(2))
    if online?(line, point) then
      return true
    end
    line.set_point(0, tri.get_point(2))
    line.set_point(1, tri.get_point(0))
    if online?(line, point) then
      return true
    end
    return false
  end

  def collision?(tri1, tri2)
    line1 = PLine.new
    line2 = PLine.new
    line3 = PLine.new
    line4 = PLine.new
    line1.set_point(0, tri1.get_point(0))
    line1.set_point(1, tri1.get_point(1))
    line2.set_point(0, tri2.get_point(0))
    line2.set_point(1, tri2.get_point(1))
    if cross?(line1, line2) then
      return true
    end
    line3.set_point(0, tri2.get_point(1))
    line3.set_point(1, tri2.get_point(2))
    if cross?(line1, line3) then
      return true
    end
    line4.set_point(0, tri2.get_point(2))
    line4.set_point(1, tri2.get_point(0))
    if cross?(line1, line4) then
      return true
    end
    line1.set_point(0, tri1.get_point(1))
    line1.set_point(1, tri1.get_point(2))
    if cross?(line1, line2) then
      return true
    end
    if cross?(line1, line3) then
      return true
    end
    if cross?(line1, line4) then
      return true
    end
    line1.set_point(0, tri1.get_point(2))
    line1.set_point(1, tri1.get_point(0))
    if cross?(line1, line2) then
      return true
    end
    if cross?(line1, line3) then
      return true
    end
    if cross?(line1, line4) then
      return true
    end
    return false
  end

  # -----------------------------------------
  # divide work
  def divide_crossline(tri2)
    div1 = []
    div2 = []
    # -- include
    if include?(@tri1, tri2) then
      return true
    end 
    if include?(tri2, @tri1) then
      return false
    end 
    # -- not collision
    if collision?(@tri1, tri2) == false then
      @div2 << tri2
      return true
    end

    # --- divide work
    # -- is tri1 point in tri2?
    if subdivision_nest(tri2) then
      @div1.clear()
      subdivision
      return true
    end

    # -- is tri2 point on tri1 edge?
    if cut(tri2) then
      @div1.clear()
      subdivision
      return true
    end

    # -- normal marge
    if clip(tri2) == false then
      @div2 << tri2
    end
    @div1.clear()
    subdivision

    # -- next
    @div1 += div1
    @div2 += div2
    return true
  end

  def subdivision_nest(tri2)
    includepoint = false
    for i in 0...3 do
      t1point = @tri1.get_point(i)
      if tri2.include?(t1point) &&
         tri2.get_point(0).equal?(t1point)==false &&
         tri2.get_point(1).equal?(t1point)==false &&
         tri2.get_point(2).equal?(t1point)==false then
        includepoint = true
        # - do nesting
        for j in 0...3 do
          tri = PTriangle.new
          tri.set_point(0, t1point)
          tri.set_point(1, tri2.get_point(j%3))
          tri.set_point(2, tri2.get_point((j+1)%3))
          if include?(@tri1, tri) == false then
            divide_crossline(tri)
          end
        end
        break
      end
    end
    return includepoint
  end

  def cut(tri2)
    onpoint = false
    #
    for i in 0...3 do
      t2point = tri2.get_point(i)
      for j in 0...3 do
        line1 = PLine.new
        line1.set_point(0, @tri1.get_point(j%3))
        line1.set_point(1, @tri1.get_point((j+1)%3))
        if online?(line1, t2point) then
          line2 = PLine.new
          line2.set_point(0, tri2.get_point((i+1)%3))
          line2.set_point(1, tri2.get_point((i+2)%3))
          if cross?(line1, line2) then
            p = crosspoint(line1, line2)
            tri = PTriangle.new
            tri.set_point(0, p)
            tri.set_point(1, t2point)
            tri.set_point(2, tri2.get_point((i+2)%3))
            tri.normalize
            divide_crossline(tri)
            tri = PTriangle.new
            tri.set_point(0, tri2.get_point((i+1)%3))
            tri.set_point(1, t2point)
            tri.set_point(2, p)
            tri.normalize
            divide_crossline(tri)
            #
            case j
            when 0
              tri1_divpoint(@epoint1, t2point)
              tri1_divpoint(@epoint1, p)
            when 1
              tri1_divpoint(@epoint2, t2point)
              tri1_divpoint(@epoint2, p)
            when 2
              tri1_divpoint(@epoint3, t2point)
              tri1_divpoint(@epoint3, p)
            end
            onpoint = true
            break
          end
        end
      end
      break if onpoint
    end
    return onpoint
  end

  def clip(tri2)
    clipping_flag = false
    for i in 0...3 do
      for j in 0...3 do
        cutter = PLine.new
        cutter.set_point(0, @tri1.get_point(j))
        cutter.set_point(1, @tri1.get_point((j+1)%3))
        e1 = PLine.new
        e1.set_point(0, tri2.get_point(i))
        e1.set_point(1, tri2.get_point((i+1)%3))
        e2 = PLine.new
        e2.set_point(0, tri2.get_point((i+1)%3))
        e2.set_point(1, tri2.get_point((i+2)%3))
        p1 = tri2.get_point((i+1)%3)
        #
        next if (cross?(cutter, e1) == false) || (cross?(cutter, e2) == false)
        #
        p2 = crosspoint(cutter, e2)
        p3 = crosspoint(cutter, e1)
        p4 = tri2.get_point(i)
        p5 = tri2.get_point((i+2)%3)
        ccheck1 = PLine.new
        ccheck1.set_point(0, p1)
        ccheck1.set_point(1, p2)
        ccheck2 = PLine.new
        ccheck2.set_point(0, @tri1.get_point(0))
        ccheck2.set_point(1, @tri1.get_point(1))
        ccheck3 = PLine.new
        ccheck3.set_point(0, @tri1.get_point(1))
        ccheck3.set_point(1, @tri1.get_point(2))
        ccheck4 = PLine.new
        ccheck4.set_point(0, @tri1.get_point(2))
        ccheck4.set_point(1, @tri1.get_point(0))
        if @tri1.include?(p1)||
           cross?(ccheck1, ccheck2) ||
           cross?(ccheck1, ccheck3) ||
           cross?(ccheck1, ccheck4) then
          cutout1 = PTriangle.new
          cutout1.set_point(0, p2)
          cutout1.set_point(1, p3)
          cutout1.set_point(2, p4)
          cutout1.normalize()
          @div2 << cutout1
          cutout2 = PTriangle.new
          cutout2.set_point(0, p4)
          cutout2.set_point(1, p5)
          cutout2.set_point(2, p2)
          cutout2.normalize()
          @div2 << cutout2
          clipping_flag = true
        else
          cutout = PTriangle.new
          cutout.set_point(0, p1)
          cutout.set_point(1, p2)
          cutout.set_point(2, p3)
          @div2 << cutout
          clipping_flag = true
        end
        #
        case j
        when 0
          tri1_divpoint(@epoint1, p2)
          tri1_divpoint(@epoint1, p3)
        when 1
          tri1_divpoint(@epoint2, p2)
          tri1_divpoint(@epoint2, p3)
        when 2
          tri1_divpoint(@epoint3, p2)
          tri1_divpoint(@epoint3, p3)
        end
      end
    end
    return clipping_flag
  end

  def subdivision
    if (@epoint1.size == 0) &&
       (@epoint2.size == 0) &&
       (@epoint3.size == 0) then
      @div1 << @tri1
      return
    end
    epoint = []
    epoint << @tri1.get_point(0)
    if @epoint1.size > 1 then
      tpoint = @epoint1.sort do |a,b|
        va = PPoint.new(a.x-@tri1.get_point(0).x,
                        a.y-@tri1.get_point(0).y,
                        a.z-@tri1.get_point(0).z)
        vb = PPoint.new(b.x-@tri1.get_point(0).x,
                        b.y-@tri1.get_point(0).y,
                        b.z-@tri1.get_point(0).z)
        va.length <=> vb.length
      end
    else
      tpoint = @epoint1
    end
    tpoint.each do |tp|
      r = false
      epoint.each do |al|
        if tp.equal?(al) then
          r = true
          break
        end
      end
      epoint << tp if r==false
    end
    epoint << @tri1.get_point(1)
    if @epoint2.size > 1 then
      tpoint = @epoint2.sort do |a,b|
        va = PPoint.new(a.x-@tri1.get_point(1).x,
                        a.y-@tri1.get_point(1).y,
                        a.z-@tri1.get_point(1).z)
        vb = PPoint.new(b.x-@tri1.get_point(1).x,
                        b.y-@tri1.get_point(1).y,
                        b.z-@tri1.get_point(1).z)
        va.length <=> vb.length
      end
    else
      tpoint = @epoint2
    end
    tpoint.each do |tp|
      r = false
      epoint.each do |al|
        if tp.equal?(al) then
          r = true
          break
        end
      end
      epoint << tp if r==false
    end
    epoint << @tri1.get_point(2)
    if @epoint3.size > 1 then
      tpoint = @epoint3.sort do |a,b|
        va = PPoint.new(a.x-@tri1.get_point(2).x,
                        a.y-@tri1.get_point(2).y,
                        a.z-@tri1.get_point(2).z)
        vb = PPoint.new(b.x-@tri1.get_point(2).x,
                        b.y-@tri1.get_point(2).y,
                        b.z-@tri1.get_point(2).z)
        va.length <=> vb.length
      end
    else
      tpoint = @epoint3
    end
    tpoint.each do |tp|
      tri1_divpoint(epoint, tp)
    end
    #
    center = PPoint.new
    center.x = (@tri1.get_point(0).x + @tri1.get_point(1).x + @tri1.get_point(2).x)/3
    center.y = (@tri1.get_point(0).y + @tri1.get_point(1).y + @tri1.get_point(2).y)/3
    center.z = (@tri1.get_point(0).z + @tri1.get_point(1).z + @tri1.get_point(2).z)/3
    #
    for i in 0...epoint.size
      at = PTriangle.new
      at.set_point(0, center)
      at.set_point(1, epoint[i])
      at.set_point(2, epoint[(i+1)%epoint.size])
      at.normalize
      @div1 << at
    end
  end


  def tri1_divpoint(memo, point)
    return if @tri1.get_point(0).equal?(point)
    return if @tri1.get_point(1).equal?(point)
    return if @tri1.get_point(2).equal?(point)
    result = false
    memo.each do |mp|
      if mp.equal?(point) then
        result = true
      end
    end
    memo << point if result == false
  end

end


# --------------------------------------------------------
#  pattern stored class
# --------------------------------------------------------

class PatternStorage

  def initialize
    @object = []
    @outline = []
  end

  attr_reader :object
  attr_reader :outline

  def object_size?
    @object.size
  end

  def clear
    @object.clear
    @outline.clear
  end

  def add( triangle )
    triangle.normalize()
    if @object.size == 0 then
      @object << triangle
    else
      marge( triangle )
    end
  end

  def drip_outline()
    @outline = []
    make_outline()
  end

  # ------------------------------
  # --- object marge work
  def marge( addtri )
    @delete_index = []
    count = @object.size
    already = false
    # -- add
    margework = TriangleCollision.new(addtri)
    for i in 0...count
      if margework.marge(@object[i]) == true then
        @delete_index << i
        @object += margework.div2
      else
        already = true
        break
      end
    end
    if !already then
      margework.build_div1()
      @object += margework.div1
    end
    # -- delete
    @delete_index.sort.reverse.each do |index|
      @object.delete_at(index)
    end
  end

  # ------------------------------
  # --- extruct object outline
  class PLineCount < PLine
    def initialize
      super
      @count = 0
    end

    attr_accessor :count
  end

  def make_outline()
    @outline = []
    linelist = []
    @object.each do |tri|
      for i in 0...3
        p1 = tri.get_point(i)
        p2 = tri.get_point((i+1)%3)
        line = PLineCount.new
        line.set_point(0, p1)
        line.set_point(1, p2)
        check = false
        linelist.each do |ol|
          if ol.same?(line) then
            ol.count += 1
            check = true
            break
          end
        end
        linelist << line if check==false
      end
    end
    # - pickup outline
    linelist.each do |ol|
      if ol.count == 0 then
        line = PLine.new
        line.set_point(0, ol.get_point(0))
        line.set_point(1, ol.get_point(1))
        @outline << line
      end
    end
  end

end



# --------------------------------------------------------
#  draw pattern primitive
# --------------------------------------------------------

class PatternObject
  include LineCrossCheck

  def initialize
    @pattern = PatternStorage.new
    @mil_path = PatternStorage.new
    @mil_outline = []
    @sequence = []
    @mil_size = 0.0
    @pad_size = 0.5
    @flash_pattern = 0
    @circle_resolution = 4
  end

  attr_accessor :circle_resolution

  def clear_pattern
    @pattern.clear
    @mil_path.clear
  end

  def clear_sequence
    @sequence.clear
  end

  def set_mil_size(size)
    @mil_size = size
  end

  def set_pad_size(size)
    @pad_size = size
  end

  def set_pad_round(type)
    if type == true then
      # round pattern
      @flash_pattern = 1
    else
      # square pattern
      @flash_pattern = 0
    end
  end

  def get_pattern_object
    @pattern.object
  end

  def get_mil_path
    @mil_outline
  end

  def set_mil_path(lines)
    @mil_outline = lines
  end

  def drip_mil_path
    @mil_path.drip_outline()
    @mil_outline = @mil_path.outline
  end

  # -----------------------------
  # --- do plot

  def flash(x, y)
    if @flash_pattern == 1 then
      draw_circle_plot(x, y, x, y, @pad_size/2, @pattern)
      draw_circle_plot(x, y, x, y, (@pad_size+@mil_size)/2, @mil_path)
    else
      draw_cube_plot(x, y, x, y, @pad_size/2, @pattern)
      draw_cube_plot(x, y, x, y, (@pad_size+@mil_size)/2, @mil_path)
    end
  end

  def move(x1, y1, x2, y2)
    if @flash_pattern == 1 then
      draw_circle_plot(x1, y1, x2, y2, @pad_size/2, @pattern)
      draw_circle_plot(x1, y1, x2, y2, (@pad_size+@mil_size)/2, @mil_path)
    else
      draw_cube_plot(x1, y1, x2, y2, @pad_size/2, @pattern)
      draw_cube_plot(x1, y1, x2, y2, (@pad_size+@mil_size)/2, @mil_path)
    end
  end


  # -----------------------------
  # --- plot line

  def get_side(x1, y1, x2, y2, size)
    # --- line angle
    if (x1 != x2) || (y1 != y2) then
      line_vector = PPoint.new(x2-x1, y2-y1, 0)
    else
      line_vector = PPoint.new(1.0, 0, 0)
    end
    line_vector.normalize
    line_vector.zoom(size)
    line_vector.rotate_z(-90.0)
    return line_vector
  end

  def draw_line(x1, y1, x2, y2, size, storage)
    width = get_side(x1, y1, x2, y2, size)
    # --- box1
    p1 = PPoint.new(x1, y1, 0)
    p2 = PPoint.new(x2, y2, 0)
    p3 = PPoint.new(x2, y2, 0)
    p4 = PPoint.new(x1, y1, 0)
    p3.add(width)
    p4.add(width)
    tri1 = PTriangle.new
    tri1.set_point(0, p1)
    tri1.set_point(1, p2)
    tri1.set_point(2, p3)
    tri1.normalize
    tri2 = PTriangle.new
    tri2.set_point(0, p3)
    tri2.set_point(1, p4)
    tri2.set_point(2, p1)
    tri2.normalize
    storage.add(tri1)
    storage.add(tri2)
    # --- box2
    width2 = PPoint.new(width.x, width.y, width.z)
    width2.rotate_z(180.0)
    p1 = PPoint.new(x1, y1, 0)
    p2 = PPoint.new(x2, y2, 0)
    p3 = PPoint.new(x2, y2, 0)
    p4 = PPoint.new(x1, y1, 0)
    p3.add(width2)
    p4.add(width2)
    tri1 = PTriangle.new
    tri1.set_point(0, p1)
    tri1.set_point(1, p2)
    tri1.set_point(2, p4)
    tri1.normalize
    tri2 = PTriangle.new
    tri2.set_point(0, p3)
    tri2.set_point(1, p4)
    tri2.set_point(2, p2)
    tri2.normalize
    storage.add(tri1)
    storage.add(tri2)
    # --- box2
  end

  def draw_cube_plot(x1, y1, x2, y2, size, storage)
    width = get_side(x1, y1, x2, y2, size)
    width.rotate_z(-90.0)
    p1 = PPoint.new(x1, y1, 0)
    p1.add(width)
    draw_line(x1, y1, p1.x, p1.y, size, storage)
    width.rotate_z(180.0)
    p2 = PPoint.new(x2, y2, 0)
    p2.add(width)
    draw_line(x2, y2, p2.x, p2.y, size, storage)
    #
    if (x1 != x2) || (y1 != y2) then
      draw_line(x1, y1, x2, y2, size, storage)
    end
  end

  def draw_circle_plot(x1, y1, x2, y2, size, storage)
    div = @circle_resolution
    width1 = get_side(x1, y1, x2, y2, size)
    width2 = PPoint.new(width1.x, width1.y, width1.z)
    width2.rotate_z(180.0)
    dangle = -(180.0 / div)
    #
    mp = PPoint.new(width1.x, width1.y, width1.z)
    pp = PPoint.new(width1.x, width1.y, width1.z)
    zp = PPoint.new(x1, y1, 0)
    for i in 0...(div-1)
      mp.rotate_z(dangle)
      p1 = PPoint.new(mp.x, mp.y, mp.z)
      p2 = PPoint.new(pp.x, pp.y, pp.z)
      p1.add(zp)
      p2.add(zp)
      tri = PTriangle.new
      tri.set_point(0, zp)
      tri.set_point(1, p1)
      tri.set_point(2, p2)
      tri.normalize
      storage.add(tri)
      pp = PPoint.new(mp.x, mp.y, mp.z)
    end
    p1 = PPoint.new(mp.x, mp.y, mp.z)
    p2 = PPoint.new(width2.x, width2.y, width2.z)
    p1.add(zp)
    p2.add(zp)
    tri = PTriangle.new
    tri.set_point(0, zp)
    tri.set_point(1, p1)
    tri.set_point(2, p2)
    tri.normalize
    storage.add(tri)
    #
    mp = PPoint.new(width2.x, width2.y, width2.z)
    pp = PPoint.new(width2.x, width2.y, width2.z)
    zp = PPoint.new(x2, y2, 0)
    for i in 0...(div-1)
      mp.rotate_z(dangle)
      p1 = PPoint.new(mp.x, mp.y, mp.z)
      p2 = PPoint.new(pp.x, pp.y, pp.z)
      p1.add(zp)
      p2.add(zp)
      tri = PTriangle.new
      tri.set_point(0, zp)
      tri.set_point(1, p1)
      tri.set_point(2, p2)
      tri.normalize
      storage.add(tri)
      pp = PPoint.new(mp.x, mp.y, mp.z)
    end
    p1 = PPoint.new(mp.x, mp.y, mp.z)
    p2 = PPoint.new(width1.x, width1.y, width1.z)
    p1.add(zp)
    p2.add(zp)
    tri = PTriangle.new
    tri.set_point(0, zp)
    tri.set_point(1, p1)
    tri.set_point(2, p2)
    tri.normalize
    storage.add(tri)
    #
    if (x1 != x2) || (y1 != y2) then
      draw_line(x1, y1, x2, y2, size, storage)
    end
  end

  # -----------------------------
  # --- marge pattern outline

  def marge_outline(item, master)
    if (cross_line?(item) == false) && (item_include?(item) == false) then
      return item
    end
    # --- keep same line
    protect_line = []
    item.each do |s_line|
      @mil_outline.each do |m_line|
        if m_line.same?(s_line) then
          protect_line << s_line
          break
        end
      end
    end
    # --- predrop include line
    dropitem = []
    item.each do |line|
      if point_include?(line) then
        dropitem << line
      end
    end
    item -= dropitem
    # --- clip cross line
    chunk = []
    dropitem = []
    item.each do |line1|
      s = split_line(line1)
      if s.size > 1 then
        dropitem << line1
        chunk += s
      end
    end
    item -= dropitem
    item += chunk
    # --- drop include line
    dropitem = []
    item.each do |line|
      if point_include?(line) then
        dropitem << line
      end
    end
    item -= dropitem
    # --- protect for master outline
    if master == true then
      item += protect_line
    else
      item -= protect_line
    end
    #
    return item
  end

  def cross_line?(item)
    result = false
    item.each do |line1|
      @mil_path.object.each do |tri|
        line2 = PLine.new
        line2.set_point(0, tri.get_point(0))
        line2.set_point(1, tri.get_point(1))
        if cross?(line1, line2) then
          result = true
        end
        line2.set_point(0, tri.get_point(1))
        line2.set_point(1, tri.get_point(2))
        if cross?(line1, line2) then
          result = true
        end
        line2.set_point(0, tri.get_point(2))
        line2.set_point(1, tri.get_point(0))
        if cross?(line1, line2) then
          result = true
        end
        if (tri[0] == line1[0]) || (tri[1] == line1[0]) || (tri[2] == line1[0]) then
          result = true
        end
        if (tri[0] == line1[1]) || (tri[1] == line1[1]) || (tri[2] == line1[1]) then
          result = true
        end
      end
      break if result
    end
    return result
  end

  def item_include?(item)
    result = false
    item.each do |line|
      if point_include?(line) then
        result = true
        break
      end
    end
    return result
  end

  def point_include?(line)
    result1 = false
    result2 = false
    @mil_path.object.each do |tri|
      if tri.include?(line.get_point(0)) then
        result1 = true
      end
      if tri.include?(line.get_point(1)) then
        result2 = true
      end
      l = PLine.new
      l.set_point(0, tri.get_point(0))
      l.set_point(1, tri.get_point(1))
      if online?(l, line.get_point(0)) then
        result1 = true
      end
      if online?(l, line.get_point(1)) then
        result2 = true
      end
      l.set_point(0, tri.get_point(1))
      l.set_point(1, tri.get_point(2))
      if online?(l, line.get_point(0)) then
        result1 = true
      end
      if online?(l, line.get_point(1)) then
        result2 = true
      end
      l.set_point(0, tri.get_point(2))
      l.set_point(1, tri.get_point(0))
      if online?(l, line.get_point(0)) then
        result1 = true
      end
      if online?(l, line.get_point(1)) then
        result2 = true
      end
      if (tri[0]==line[0])||(tri[1]==line[0])||(tri[2]==line[0]) then
        result1 = true
      end
      if (tri[0]==line[1])||(tri[1]==line[1])||(tri[2]==line[1]) then
        result2 = true
      end
    end
    return (result1==true && result2==true)
  end

  def cutting(line1, line2)
    clip = []
    xp = crosspoint(line1, line2)
    nl1 = PLine.new
    nl1.set_point(0, line1.get_point(0))
    nl1.set_point(1, xp)
    nl2 = PLine.new
    nl2.set_point(0, xp)
    nl2.set_point(1, line1.get_point(1))
    clip << nl1
    clip << nl2
    return clip
  end

  def split_line(line)
    chunk = []
    @mil_path.object.each do |tri|
      line1 = PLine.new
      line1.set_point(0, tri.get_point(0))
      line1.set_point(1, tri.get_point(1))
      line2 = PLine.new
      line2.set_point(0, tri.get_point(1))
      line2.set_point(1, tri.get_point(2))
      line3 = PLine.new
      line3.set_point(0, tri.get_point(2))
      line3.set_point(1, tri.get_point(0))
      if cross?(line, line1) then
        splitline = cutting(line, line1)
        chunk += split_line(splitline[0])
        chunk += split_line(splitline[1])
        break
      elsif cross?(line, line2) then
        splitline = cutting(line, line2)
        chunk += split_line(splitline[0])
        chunk += split_line(splitline[1])
        break
      elsif cross?(line, line3) then
        splitline = cutting(line, line3)
        chunk += split_line(splitline[0])
        chunk += split_line(splitline[1])
        break
      end
    end
    if chunk.size == 0 then
      chunk << line
    end
    return chunk
  end

end

